package edu.udo.cs.ie.cowbots.arch;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import edu.udo.cs.ie.cowbots.arch.massim.MassimAgent;
import jason.architecture.AgArch;
import jason.asSyntax.Literal;

/**
 * Implementation of EnvionmentAdapter to connect the 
 * Cowbot MAS to Massim. It is used to receive perceptions
 * from massim an convert them to literals used for
 * the Cowbots beliefs, and to send the cowbots
 * actions to massim so they can interact with the
 * environment they 'live' in.
 * 
 * @author dhoelzgen
 *
 */
public class MassimAdapter extends EnvironmentAdapter {

	/**
	 * Instance of an MassimAgent, wich takes part in the
	 * massim simulation.
	 */
	private MassimAgent massimAgent;
	
	// jason name of agent
	private String	agName;
	
	// current world time
	private int	worldTime;
	
	// logger
	Logger log;
	
	/**
	 * Creates a new instance of an massim adapter
	 * 
	 * @param owner AgentArch, to wake up the reasoning cycle with
	 * given new perceptions
	 * @param host The host massim is running on
	 * @param username The agents username
	 * @param pass The agents password
	 * @param port The port massim is listening to
	 */
	public MassimAdapter(AgArch owner, String host, String username, String pass, int port) {
		super(owner);
		agName = owner.getAgName();
		log = Logger.getLogger(agName);
		log.info("massim adapter initialized");
		worldTime = 0;
		massimAgent = new MassimAgent(this, host, username, pass, port);
		massimAgent.start();
	}
	
	/**
	 * Receives the perceptions from massim in xml format, and
	 * converts them to literals used for the Cowbots beliefs.
	 * 
	 * @param percepts XML Root Element
	 */
	public void processMassimPercepts(Element percepts) {
		List<Literal> l = new ArrayList<Literal>();
		Element root = percepts;
		
		String step = root.getAttribute("step");
		
		// fix: shift worldtime by 1
		int istep = Integer.parseInt(step.trim());
		++istep;
		step = ""+istep;
		this.worldTime = istep;
		
		int agX = Integer.parseInt(root.getAttribute("posx"));
		int agY = Integer.parseInt(root.getAttribute("posy"));
		
		// fix:
		// create current world time
		l.add( Literal.parseLiteral("time("+step+")" ));
		//l.add( addPerceptWithIDAndPosition(agName, agX, agY, step) );
		
		log.info("massim percepts: ("+agX+","+agY+"), step "+step);
		
		try {
			NodeList nl = root.getElementsByTagName("cell");

			for (int i = 0; i < nl.getLength(); i++) {
				Node n = nl.item(i);
				Element ne = (Element) n;
				NodeList nnl = n.getChildNodes();

				String content = nnl.item(1).getNodeName();
				int directionX = Integer.parseInt(ne.getAttribute("x"));
				int directionY = Integer.parseInt(ne.getAttribute("y"));

				int x = massimAgent.posX + directionX;
				int y = massimAgent.posY + directionY;

				if (content.equalsIgnoreCase("agent")) {
					Element nne = (Element) nnl.item(1);
					if (nne.getAttribute("type").equalsIgnoreCase("ally")) {
						l.add(addPerceptWithIDAndPosition("ally", x, y, step));
						l.add(addPerceptIdentificator("ally","ally"));
					} else {
						l.add(addPerceptWithIDAndPosition("enemy", x, y, step));
						l.add(addPerceptIdentificator("enemy","enemy"));
					}
				}

				if (content.equalsIgnoreCase("obstacle")) {
					l.add(addPerceptWithIDAndPosition("obstacle", x, y, step));
					l.add(addPerceptIdentificator("obstacle","obstacle"));
				}

				if (content.equalsIgnoreCase("cow")) {
					Element nne = (Element) nnl.item(1);
					String cowId = nne.getAttribute("ID");
					l.add(Literal.parseLiteral("cow(cow" + cowId + ")"));
					l.add(addPerceptWithIDAndPosition("cow" + cowId, x, y, step));
				}

				if (content.equalsIgnoreCase("empty")) {
					l.add(addPerceptWithIDAndPosition("empty", x, y, step));
					l.add(addPerceptIdentificator("empty","empty"));
				}

				if (content.equalsIgnoreCase("corral")) {
					Element nne = (Element) nnl.item(1);
					if (nne.getAttribute("type").equalsIgnoreCase("ally")) {
						l.add(addPerceptWithIDAndPosition("owncorral", x, y, step));
						l.add(addPerceptIdentificator("owncorral","owncorral"));
					} else {
						l.add(addPerceptWithIDAndPosition("enemycorral", x, y, step));
					}
				}

				if (content.equalsIgnoreCase("switch")) {
					//tv: add every switch as "switch_X_Y"
					String switchname = "switch_"+x+"_"+y;
					l.add(addPerceptWithIDAndPosition(switchname, x, y, step));					
					//l.add(addPerceptIdentificator("switch","switch"));
					l.add(addPerceptIdentificator("switch",switchname));
				}

				if (content.equalsIgnoreCase("fence")) {
					Element nne = (Element) nnl.item(1);
					l.add(addPerceptWithIDAndPosition("fence", x, y, step));
					l.add(addPerceptIdentificator("fence","fence"));
					if (nne.getAttribute("open").equalsIgnoreCase("true"))
						l.add(Literal.parseLiteral("isopen(" + x + "," + y + "," + "true," + step + ")"));
					else
						l.add(Literal.parseLiteral("isopen(" + x + "," + y + "," + "false," + step + ")"));
				}

			}

			//l.add(addPerceptWithIDAndPosition("myself", massimAgent.posX, massimAgent.posY, step));
			l.add( addPerceptWithIDAndPosition(agName, agX, agY, step) );
			
		} catch (Exception ex) {
			System.out.println("Waiting for Server...");
		}
		
		this.newPercepts(l);
	}

	/**
	 * Called when simulation starts, receives some information
	 * about the world size. It is used to generate beliefs for every
	 * unexplored field of the world map and to generate beliefs about
	 * the world size.
	 * 
	 * @param percepts XML Root Element
	 */
	public void processMassimSimStart(Element percepts) {
		List<Literal> l = new ArrayList<Literal>();

		int mapWidth = Integer.valueOf(percepts.getAttributeNode("gsizex").getNodeValue());
		int mapHeight = Integer.valueOf(percepts.getAttributeNode("gsizey").getNodeValue());
		String simId = percepts.getAttributeNode("id").getNodeValue();
		int los = Integer.valueOf(percepts.getAttributeNode("lineOfSight").getNodeValue());
		// fix:
		// add worldsize (very important for us)
		
		// 0
		l.add( Literal.parseLiteral("simulation("+simId+")"));
		
		// 1
		l.add( Literal.parseLiteral( "worldsize(" + mapWidth + "," + mapHeight + ")" ) ); 
		
		// 2
		l.add( Literal.parseLiteral( "lineofsight("+los+")" ));
		
		// 3
		l.add(addPerceptIdentificator("owncorral","owncorral"));
		
		// 4
		l.add(addPerceptIdentificator("unexplored","unexplored"));
		
		// fix:
		// parse corral coordinates
		int cMinX = Integer.valueOf(percepts.getAttributeNode("corralx0").getNodeValue());
		int cMinY = Integer.valueOf(percepts.getAttributeNode("corraly0").getNodeValue());
		int cMaxX = Integer.valueOf(percepts.getAttributeNode("corralx1").getNodeValue());
		int cMaxY = Integer.valueOf(percepts.getAttributeNode("corraly1").getNodeValue());
		
		int corCX = cMinX + (cMaxX-cMinX)/2;
		int corCY = cMinY + (cMaxY-cMinY)/2;
		
		// 5
		l.add( Literal.parseLiteral( "corralcenter("+corCX+","+corCY+")" ));
		
		// 6
		l.add( Literal.parseLiteral("corralpos("+cMinX+","+cMinY+","+cMaxX+","+cMaxY+")"));
		
		/*
		for (int i = 0; i < mapWidth; i++) {
			for (int j = 0; j < mapHeight; j++) {
				if ((i >= cMinX) && (i<=cMaxX) && (j >= cMinY) && (j <= cMaxY))
					l.add(this.addPerceptWithIDAndPosition("owncorral", i, j, "0"));
				else
					l.add(this.addPerceptWithIDAndPosition("unexplored", i, j, "0"));
			}
		}
		*/
		
		
		this.newSimulation(l);
	}
	
	
	private Literal addPerceptIdentificator(String ident, String id) {
		return Literal.parseLiteral(ident+"("+id+")");
	}
	
	/**
	 * Utility functions for generating default 'ison' literals
	 * 
	 * @param name
	 * @param x
	 * @param y
	 * @param step
	 * @return
	 */
	private Literal addPerceptWithIDAndPosition(String name, int x, int y, String step) {
		return Literal.parseLiteral("ison(" + name + ", " + x + "," + y + ", " + step + ")");
	}
	
	/**
	 * Method so set the next action performed by the
	 * agent in the massim environment.
	 * 
	 * @param action Action performed by the agent
	 * Possible actions are: skip, north, south, east, west,
	 * northwest, northeast, southwest, southeast
	 */
	@Override
	public void act(String action) {
		massimAgent.setNextAction(action);
	}

	@Override
	public int getWorldTime() {
		return this.worldTime;
	}

	public void processSimulationEnd(Element perception, long currenttime) {
		List<Literal> l = new ArrayList<Literal>();
		l.add( Literal.parseLiteral("endOfSimulation"));
		this.newEndOfSimulation(l);
	}
}
