/**
 * 
 */
package edu.udo.cs.ie.cowbots.bdi.brf;

import jason.asSyntax.Literal;
import jason.asSemantics.Message;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.logic.solver.*;
import java.util.*;

/**
 * This class holds the belief set from the Cowbot-BDI-model. Beside
 * simple set-container functionality, the cowbot belief set allows
 * to search for beliefs that matches a given predicate symbol.
 * 
 * @author Boguslaw, Thomas
 * 
 */
public class BeliefSet {
	
	protected	Set<ELPLiteral> bels;	
	protected	IdentTable 	idents;
	protected	Map<String,List<ELPLiteral>> indexedBels;
	
	/**
	 * constructs a new belief set with given beliefs
	 * 
	 * @param bel	initial beliefs
	 */
	public BeliefSet(Set<ELPLiteral> bel, IdentTable ids) {
		this.bels = bel;
		if (this.bels == null)
			this.bels = new HashSet<ELPLiteral>();
		this.idents = ids;
		
		// create index table
		this.indexBeliefs();
	}
	
	/**
	 * auxiliary function that indexes all beliefs by their
	 * predicate symbol
	 */
	protected void indexBeliefs() {
		this.indexedBels = new HashMap<String, List<ELPLiteral>>();
		for (ELPLiteral l : this.bels) {
			String key = l.getAtom().getPredicate();
			
			List<ELPLiteral> values = indexedBels.get(key);
			if (values == null) {
				values = new LinkedList<ELPLiteral>();
				indexedBels.put(key, values);
			}
			values.add(l);
		}
	}
	
	/**
	 * 
	 * @return	set of beliefs
	 */
	public Set<ELPLiteral>  getBeliefs() {
		return bels;
	}
	
	@Override
	public String toString() {
		return "belief set, " + bels.size() + " beliefs";
	}
	
	
	/**
	 * returns all beliefs that starts with a predicate
	 * @param predicate	functor name to look for
	 * @return	list of all literals matching predicate name
	 */
	public List<ELPLiteral> getBeliefsByPredicate(String predicate) {
		List<ELPLiteral> ret = new LinkedList<ELPLiteral>();
		for (ELPLiteral l : this.bels) {
			if (l.getAtom().getPredicate().equals(predicate) )
				ret.add(l);
		}
		return ret;
	}
	
	/**
	 * removes a literal from the belief set
	 * @param l	literal to delete
	 */
	public void deleteBelief(ELPLiteral l) {
		if (l != null) {
			if (this.bels.remove( l )) {
				String index = l.getAtom().getPredicate();
				Iterator<ELPLiteral> iter = this.indexedBels.get(index).iterator();
				while (iter.hasNext()) {
					ELPLiteral inBels = iter.next();
					if (inBels.equals(l))
						iter.remove();
				}
			}
		}
	}
	
	/**
	 * adds a belief to the belief set (doesn't neccessary
	 * means that it was added to the belief base before,
	 * or it will survive the next revision cycle).
	 * @param l	literal to add
	 */
	public void addBelief(ELPLiteral l) {
		if (l != null)
			if (this.bels.add(l)) {
				String index = l.getAtom().getPredicate();
				List<ELPLiteral> ls = this.indexedBels.get(index);
				if (ls == null) {
					ls = new LinkedList<ELPLiteral>();
					this.indexedBels.put(index, ls);
				}
				ls.add(l);
			}
	}
	
	/**
	 * returns an IdentTable lookup instance or null (optional
	 * operation).
	 * 
	 * @return	IdentTable instance (optional or empty)
	 */
	public IdentTable getIdentifiers() {
		return this.idents;
	}
	
	/**
	 * returns the first predicate found in the belief set
	 * starting with a matching functor 
	 * @param functor literal with functor name to search for
	 * @return	null or first found literal
	 */
	public ELPLiteral findFirst(String functor) {		
		List<ELPLiteral> belList = this.indexedBels.get(functor);
		if (belList == null)
			return null;
		if (belList.size() <= 0)
			return null;
		
		ELPLiteral ret = null;
		Iterator<ELPLiteral> belIter = belList.iterator();
		while (belIter.hasNext()) {
			ELPLiteral l = belIter.next();
			if (l.getAtom().getPredicate().equalsIgnoreCase(functor)) {
				ret = l;
				break;
			}
		}
		
		return ret;
	}
	
	/**
	 * returns a list of all literals matching a given functor
	 * @param functor	predicate symbol to look for
	 * @return		list of all literals with predicate symbol <i>functor</i>
	 */
	public List<ELPLiteral> findAll(String functor) {
		List<ELPLiteral> ret = this.indexedBels.get(functor);
		if (ret != null)
			return ret;
		else
			return Collections.emptyList();
	}
}
