package edu.udo.cs.ie.cowbots.logic;

import java.util.*;
import edu.udo.cs.ie.cowbots.logic.*;

/**
 * this class models an alphabet, the set of all
 * predicate symbols appearing in an extended logic
 * program.
 * 
 * @author Thomas
 *
 */
public class Alphabet {

	Set<String>		headSet, bodySet,sharedSet, allSet;
	
	/**
	 * default constructor, results in an
	 * empty alphabet
	 */
	public Alphabet() {
		headSet = new HashSet<String>();
		bodySet = new HashSet<String>();
		sharedSet = new HashSet<String>();
		allSet = new HashSet<String>();
	}
	
	/**
	 * constructs an alphabet based on a given program 
	 * @param program initial program to build alphabet from
	 */
	public Alphabet(ELP program) {
		this();
		addSymbols(program);
	}
	
	/**
	 * adds all functors symbols in a given program to the alphabet
	 * @param program logical program to add symbols from
	 */
	public void addSymbols(ELP program) {
		for (ELPRule r : program.getRules()) {
			for (int i = 0; i < r.nHead(); i++)
				this.addHeadSymbol( r.getHead(i).getAtom().getPredicate());
			for (int i = 0; i < r.nBody(); i++)
				if (r.getBody(i).getAtom().isPredicate())
					continue;
				else
					this.addBodySymbol( r.getBody(i).getAtom().getPredicate());
					
		}
	}
	
	/**
	 * returns the set of all functor names
	 * that appear at least in a head of a program rule
	 * @return set of head symbols
	 */
	public Set<String> getHeadSymbols() {
		return headSet;
	}
	
	/**
	 * returns the set of all functor names
	 * that appear at least in a body of a program rule
	 * @return set of head symbols
	 */
	public Set<String> getBodySymbols() {
		return bodySet;
	}
	
	/**
	 * returns the set of all functor names
	 * that appear in a head and a body of
	 * some rules in a program
	 * @return set of intersection of head and body symbols
	 */
	public Set<String> getSharedSymbols() {
		return sharedSet;
	}
	
	/**
	 * returns the set of all functor names
	 * appearing in a program
	 * @return set of all symbols
	 */
	public Set<String> getSymbols() {
		return allSet;
	}
	
	/**
	 * adds a symbol appearing in a head of
	 * a program to the alphabet
	 * @param sym functor symbol to add
	 */
	public void addHeadSymbol(String sym) {
		headSet.add(sym);
		if (bodySet.contains(sym))
			sharedSet.add(sym);
	}
	
	/**
	 * adds a symbol appearing in a body of
	 * a program to the alphabet
	 * @param sym functor symbol to add
	 */
	public void addBodySymbol(String sym) {
		bodySet.add(sym);
		if (headSet.contains(sym))
			sharedSet.add(sym);
	}
	
	public String toString() {
		String ret = "head symbols: "+headSet.toString();
		ret += "\nbody symbols: "+bodySet.toString();
		return ret;
	}
}
