package edu.udo.cs.ie.cowbots.ui;

import javax.swing.*;

import edu.udo.cs.ie.cowbots.ui.theme.CowbotColorTable;
import edu.udo.cs.ie.cowbots.util.cluster.CowCluster;


import java.awt.*;
import java.awt.geom.*;
import java.util.concurrent.locks.*;
import java.util.*;
import java.util.List;

/**
 * a basic cell painting panel based
 * upon a java swing jpanel.
 * 
 * @author Thomas
 *
 */
public class MapPanel extends JPanel {
	
	// lock to protect internal data from race conditions
	// when multiple threads are updating or reading cell data. 
	protected ReentrantLock			paintLock;
	
	// multiple buffered layers containing cell data
	protected ArrayList<ArrayList<DrawCell>>		buffers;
	
	// collection of rectangles used by the debug output of the
	// clustering algorithm.
	protected List<CowCluster>				clusters;
	
	// a* points
	protected List<Point>				pathpoints;
	
	// waypoints 
	protected LinkedList<Point>						waypoints;
		
	// dimension of the cell world to display (can be zero)
	protected int									mapX, mapY;
	
	// autoscaling tries to resize all drawn cells to the
	// jpanels screen area.
	protected boolean								autoScale;
	
	public MapPanel() {
		super();
		paintLock = new ReentrantLock();

		buffers = new ArrayList<ArrayList<DrawCell>>();
		buffers.add(new ArrayList<DrawCell>());
		buffers.add(new ArrayList<DrawCell>());
		
		clusters = new LinkedList<CowCluster>();
		waypoints = new LinkedList<Point>();
		pathpoints = null;
		
		mapX = 1;
		mapY = 1;
		autoScale = false;
	}
	
	/**
	 * sets the first draw buffer to db
	 * @param db
	 */
	public void setDrawBuffer(ArrayList<DrawCell> db) {
		paintLock.lock();
		
		// copy draw buffer
		ArrayList<DrawCell> buffer = buffers.get(0);
		buffer.clear();
		buffer.addAll(db);
		
		paintLock.unlock();
	}
	
	/**
	 * sets a draw buffer to input
	 * 
	 * @param db		input draw commands
	 * @param layer		internal draw buffer index
	 */
	public void setDrawBuffer(ArrayList<DrawCell> db, int layer) {
		paintLock.lock();
		
		// copy draw buffer
		ArrayList<DrawCell> buffer = buffers.get(layer & 1); 
		buffer.clear();
		buffer.addAll(db);
		
		paintLock.unlock();
	}
	
	/**
	 * sets the dimension, in cells, of the draw panel. this data
	 * is used internally by some draw routines.
	 *  
	 * @param width		number of cells horizontally
	 * @param height	number of cells vertically
	 */
	public void setMapSize(int width, int height) {
		if ((width > 0) && (height > 0)) {
			this.mapX = width;
			this.mapY = height;
		}
	}
	
	/**
	 * enables or disables auto scaling. if enabled,
	 * each cell is resized that all cells covers
	 * the painting panel's screen area.
	 * 
	 * @param enable	enables (true) or disables (false) auto scaling
	 * 
	 * @author Thomas
	 */
	public void setAutoScale(boolean enable) {
		this.autoScale = true;
	}
	
	/**
	 * retrieve auto scale state
	 * 
	 * @return	state of auto scaling
	 * 
	 * @author Thomas
	 */
	public boolean getAutoScale() {
		return this.autoScale;
	}
	
	/**
	 * map panels core drawing routine
	 * 
	 * @author Thomas
	 */
	@Override
	public void paintComponent(Graphics g) {
		
		paintLock.lock();
		prepareFrame();
		paintLock.unlock();
		
		//super.paintComponent(g);
		
		// setup
		Graphics2D g2 = (Graphics2D) g;
		
		int width = this.getWidth();
		int height = this.getHeight();
		
		int cellX = 4;
		int cellY = 4;
		
		if ((autoScale) && (mapX > 0) && (mapY > 0)) {
			int bestX = width / mapX;
			int bestY = height / mapY;
			int bestXY = (bestX < bestY)? bestX:bestY;
			if (bestXY > 4) {
				cellX = bestXY;
				cellY = bestXY;
			} else {
				cellX = 8;
				cellY = 8;
			}			
		}
		
		// clear the background
		g2.setColor(Color.BLACK);
		g2.fillRect(0, 0, width, height);
		
		// process draw buffer
		// it would be nice if we could track down auto sizing
		paintLock.lock();
		
		Color col = Color.BLACK;
		for (ArrayList<DrawCell> drawBuffer : buffers)
			for (DrawCell c : drawBuffer) {
				g2.setColor(c.c);
				g2.fillRect(c.x*cellX, c.y*cellY, cellX, cellY);
			}
		
		// draw clusters
		for (CowCluster cc : clusters)
			cc.draw(g2, cellX, cellY);
		
		// draw waypoints
		g2.setColor( Color.CYAN );
		for (Point p : this.waypoints) {
			g2.drawLine( p.x*cellX, p.y*cellY, (p.x+1)*cellX, (p.y+1)*cellY);
			g2.drawLine( p.x*cellX, (p.y+1)*cellY, (p.x+1)*cellX, p.y*cellY);
		}
		
		// draw path points
		if (pathpoints != null)
			drawPathPoints(g2,cellX,cellY);
		
		paintLock.unlock();
		
		g2.dispose();
	}

	/**
	 * users can override this routine to set up local draw buffers
	 * before anything is drawn onto the underlying jpanel.
	 * 
	 * @author Thomas
	 */
	protected void prepareFrame() {
		
	}

	/**
	 * loads an image with an optional colorkey for transparency
	 * 
	 * @return
	 */
	public int loadImage() {
		return -1;
	}
	
	
	public void setClusterList(List<CowCluster> clusters) {
		this.paintLock.lock();
		this.clusters = clusters;
		this.paintLock.unlock();
	}
	
	public void setWaypoints(LinkedList<Point> wp) {
		this.paintLock.lock();
		this.waypoints.clear();
		this.waypoints.addAll(wp);
		this.paintLock.unlock();
	}
	
	public void setPathPoints(List<Point> path) {
		this.paintLock.lock();
		this.pathpoints = path;
		this.paintLock.unlock();
	}
	
	protected void drawPathPoints(Graphics2D g2, int cellX, int cellY) {
		int cellXo2 = cellX/2;
		int cellYo2 = cellY/2;
		
		Iterator<Point> pIter = pathpoints.iterator();
		
		g2.setColor(Color.CYAN);
		
		if (pIter.hasNext()) {
			Point pStart = pIter.next();
			
			while (pIter.hasNext()) {
				Point pNext = pIter.next();
				
				g2.drawLine(pStart.x*cellX+cellXo2, pStart.y*cellY+cellYo2, 
						pNext.x*cellX+cellXo2, pNext.y*cellY+cellYo2);
								
				pStart = pNext;
			}
		}
		
	}
}
