package edu.udo.cs.ie.cowbots.util;

import java.awt.Point;
import java.util.LinkedList;
import java.util.Queue;


/**
 * this class models a distance map that
 * is used to evaluate the distance between
 * cells on a tileworld.
 * 
 * @author Thomas
 *
 */
public class DistanceMap {

	int map[][];		// distance representation of the map
	int penalty[][];	// optional penalty for a field
	int attrib[][];	// attribute map (bit 0: obstacle, bit 1: at switch
	int width, height;
	int borderGuard;
	int fillX,fillY;
	
	static final byte BIT_OBSTACLE = 1;		// bit for an obstacle
	static final byte BIT_ATSWITCH = 2;		// bit for a switch cell
	
	// relative coordinates to a given cell. array order is
	// north|east|south|west | northeast|southeast|southwest|northwest
	static int[] relX = {0,1,0,-1,  1,1,-1,-1};
	static int[] relY = {1,0,-1,0,  -1,1,1,-1 };
	static String[] relDir = {"south","east","north","west","northeast",
		"southeast","southwest","northwest"};
	
	int	bestSearches[] = new int[8];
	int	penSearches[] = new int[8];

	
	/**
	 * internal used helper class for next fill
	 * point
	 * 
	 * @author Thomas
	 *
	 */
	class FillStart {
		
		public int x,y,value;
		
		public FillStart(int x, int y, int value) {
			this.x = x;
			this.y = y;
			this.value = value;
		}
	}
	
	/**
	 * creates a distance map
	 * 
	 * @param x	width of map
	 * @param y height of map
	 */
	public DistanceMap(int x, int y) {
		map = new int[y+2][x+2];
		penalty = new int[y+2][x+2];
		attrib = new int[y+2][x+2];
		
		width = x;
		height = y;
		
		borderGuard = x*x+y*y;
		
		for (int i = 0; i < x+2; i++) {
			attrib[0][i] = 1;
			attrib[y+1][i] = 1;
			map[0][i] = borderGuard;
			map[y+1][i] = borderGuard;
		}
		for (int i = 0; i < y+2; i++) {
			attrib[i][0] = 1;
			attrib[i][x+1] = 1;
			map[i][0] = borderGuard;
			map[i][x+1] = borderGuard;
		}
	}
	
	
	/**
	 * creates distances from a starting point
	 * 
	 * @param sx	start x
	 * @param sy	start y
	 */
	public void fill(int sx, int sy) {
		
		fillX = sx;
		fillY = sy;
		
		preFill();
		
		++sx;
		++sy;
		
		FillStart fs = new FillStart(sx,sy,1);
		Queue<FillStart> q = new LinkedList<FillStart>();
		
		map[fs.y][fs.x] = fs.value;
		q.add(fs);
		
		while (q.peek() != null) {
			
			fs = q.poll();
			
			//map[fs.y][fs.x] = fs.value;
			//int pen = penalty[fs.y][fs.x]; 
			int val = fs.value+1;
			
			// spawn new fill points
			int nx = fs.x +1;
			int ny = fs.y;
			if (nx <= this.width) { 
				if ((map[ny][nx] == 0) && ((attrib[ny][nx] & BIT_OBSTACLE) == 0)) {
					map[ny][nx] = val;
					q.add( new FillStart(nx,ny,val));	
				}
			}
			nx = fs.x -1;
			ny = fs.y;
			if (nx > 0) {
				if ((map[ny][nx] == 0) && ((attrib[ny][nx] & BIT_OBSTACLE) == 0)) {
					map[ny][nx] = val;
					q.add( new FillStart(nx,ny,val));	
				}
			}
			
			nx = fs.x;
			ny = fs.y+1;
			if (ny <= this.height) {
				if ((map[ny][nx] == 0) && ((attrib[ny][nx] & BIT_OBSTACLE) == 0)) {
					map[ny][nx] = val;
					q.add( new FillStart(nx,ny,val));	
				}
			}
			nx = fs.x;
			ny = fs.y-1;
			if (ny > 0) {
				if ((map[ny][nx] == 0) && ((attrib[ny][nx] & BIT_OBSTACLE) == 0)) {
					map[ny][nx] = val;
					q.add( new FillStart(nx,ny,val));	
				}
			}
		}
	}
	
	
	/**
	 * initializes the map with initial distances (0 for every cell)
	 */
	protected void preFill() {
		/*
		int yLen = map.length-1;
		int xLen = map[0].length-1;
		
		for (int j = 1; j < yLen; j++)
			for (int i = 1; i < xLen; i++)
				map[j][i] = 0;
		*/
	}
	
	/**
	 * this method returns a path to the fill position. if no
	 * path exists, null is returned. if the distance is already
	 * zero, an empty list is returned. otherwise, the next
	 * cells are returned. if no path exists, null is returned.
	 *  
	 * @param fromX	from start x
	 * @param fromY from start y
	 * @return null - no path avail or list with cells to go to
	 */
	public LinkedList<Point> getPath(int fromX, int fromY) {
				
	 	if (map[fromY][fromX] == 0)
	 		return null;
	 	
	 	LinkedList<Point> ret = new LinkedList<Point>();	 	
	 	ret.add( new Point(fromX,fromY));
	 	
	 	// special case to handle own location
	 	if ((fromX == this.fillX) && (fromY == this.fillY)) {	 		
	 		return ret;
		}
	 	
	 	
	 	// adjust points (distance map is shifted bei +1,+1 to
		// make tests at the playfield border easies)
		fromX+=1;
		fromY+=1;
		
	 	// get best field
	 	boolean goon = true;
	 	while (goon) {
	 		
	 		// detect best adjacent cell
	 		int refValue = map[fromY][fromX];
	 		
	 		int i = fromX;
	 		int j = fromY;
	 		
	 		// find best direction
	 		int fieldAttrib = attrib[j][i];
	 		
	 		int searchLen = 8; // 8 for diagonal move, 4 for axis-aligned movement only
	 		if ((fieldAttrib & this.BIT_ATSWITCH) != 0) {
	 			searchLen = 4;
	 		}
	 		
	 		int results = 0;
	 		int bestResult = -1;
	 		int bestValue = refValue;
	 		int bestResultNoPen = -1;
	 		int bestValueNoPen = refValue;
	 		
	 		for (int w = 0; w < searchLen; w++) {
	 			int tmpValue = map[j+relY[w]][i+relX[w]];
	 			int tmpPenalty = penalty[j+relY[w]][i+relX[w]];
	 			
	 			// take cells with lowest distance
	 			if (tmpValue < refValue)  {
	 				
	 				if (tmpPenalty == 0) {
	 					if (tmpValue < bestValueNoPen) {
	 						bestValueNoPen = tmpValue;
	 						bestResultNoPen = w;
	 					}
	 				}
	 				else {
	 					if (tmpValue < bestValue) {
	 						bestValue = tmpValue;
	 						bestResult = w;
	 					}
	 				}	 					 				
	 			} 
	 		}
	 		
	 		// if we have a penalty field,always try to
	 		// prefer non-penetalized field
	 		if (bestResult >= 0) {
	 			if (bestResultNoPen > -1) {
	 				bestResult = bestResultNoPen;
	 			}	 			
	 		} else {
	 			bestResult = bestResultNoPen;
	 		}
	 		
	 		if (bestResult >= 0) {	 			
	 			fromX += relX[bestResult];
	 			fromY += relY[bestResult];
	 		} else {
	 			break;
	 		}
	 		
	 		
	 		// store path
	 		ret.add( new Point( fromX-1,fromY-1 ));
	 			 		
	 	}
	 	
	 	// if the last cell is an obstacle, return no path at all!
	 	if ( (attrib[fromY][fromX] & this.BIT_OBSTACLE) != 0)
	 		return null;
	 	
	 	return ret;
	}
	
	
	/**
	 * returns the distance from a point to the
	 * distance maps reference point. the meaning
	 * of the return values is:
	 * 0  : no path exists to the fill point
	 * 1  : same field
	 * >1 : distance to fill point
	 * 
	 * @param x	point to query distance for (x)
	 * @param y point to query distance for (y)
	 * @return distance of (x,y) to reference point
	 */
	public int distance(int x, int y) {
		return map[y+1][x+1];
	}
	
	
	/**
	 * sets an attribute for a cell. a cell is a byte value,
	 * where 0 indicating an empty cell.
	 * 
	 * @param x	x coordinate part
	 * @param y y coordinate part
	 * @param attr cell attribute to set
	 */
	public void setAttrib(int x, int y, byte attr) {
		this.attrib[y+1][x+1] = attr;
		this.map[y+1][x+1] = this.borderGuard;
	}
	
	
	public void setPenalty(int x, int y, int pen) {
		this.penalty[y+1][x+1] = pen;
	}
	
	
	/**
	 * marks a cell as an obstacle (unpassable)
	 * @param x
	 * @param y
	 */
	public void setObstacle(int x, int y) {
		int obsX = x+1;
		int obsY = y+1;
		this.map[obsY][obsX] = this.borderGuard;
		this.attrib[obsY][obsX] = this.attrib[obsY][obsX] | this.BIT_OBSTACLE;
	}
	
	/**
	 * marks a cell as adjacent to a switch
	 * @param x
	 * @param y
	 */
	public void setAtSwitch(int x, int y) {
		this.attrib[y+1][x+1] = this.attrib[y+1][x+1] | this.BIT_ATSWITCH;
	}
	
	
	/**
	 * debug output to console
	 */
	public void debugPrint() {
		for (int y = 0; y < map.length; y++) {
			for (int x = 0; x < map[y].length; x++) {
				if (attrib[y][x] == 1)
					System.out.print("#");
				else
					System.out.print(i2c( map[y][x] ));
			}
			System.out.println();
		}
	}
	
	/**
	 * helper method, converts an integer to a 
	 * character between 0..9, a..z. wraps around
	 * z if input value is to high.
	 * 
	 * @param i	positive integer
	 * @return character value for input
	 */
	protected char i2c(int i) {
		char ret = '1';
		
		if ((i >= 0) && (i <= 9)) {
			ret = (""+i).charAt(0);
		}
		
		if ((i >= 10) && (i <= 35)) {
			i-=10;
			i+=97;
			
			ret = Character.toChars(i)[0];
		}
				
		return ret;
	}
}
