package edu.udo.cs.ie.cowbots.util.cluster;

import java.awt.Point;
import java.util.*;
import java.util.Map.Entry;
import edu.udo.cs.ie.cowbots.internalActions.scout.astar;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.util.DistanceMap;
import edu.udo.cs.ie.cowbots.util.GridMap;


/**
 * another cluster calculator, mainly a speed-up
 * version of the first on.
 * 
 * @author Thomas
 *
 */
public class NewClusterCalculator implements IClusterCalculator {

	private int clusterDistance = 4;

	private astar pathfinder;

	public NewClusterCalculator() {
		pathfinder = new astar();
	}

	/**
	 * this method returns a set of clusters of cows
	 * derived from the current world view of an agent.
	 * 
	 * @param map	GridMap representing an agents view of the world
	 * @param cDist	maximum allowed cluster elongation
	 * @param maxCowTime	maximum age of cows to respect during clustering
	 * @return	list of cow cluster
	 */
	public List<CowCluster> getCowClusters(GridMap map, DistanceMap dm, int cDist, int maxCowTime) {
		
		LinkedList<CowCluster> cList = new LinkedList<CowCluster>();
		
		// TODO: sort cows so that cows close together on the map
		// are also close together in the list. that should help
		// the trivial cluster search within the loop dramatically.
		
		// get cows from the gridmap
		for (ELPLiteral l : map.getMoversByType("cow")) {
			
			CowPosition cp = new CowPosition(l);
			int cowTime = l.getAtom().getTermAsInt(3);
			
			boolean clusterFound = false;
			
			// ignore all cows we we cannot get a meaningful
			// distance, or with are cornered (has two
			// adjacent obstacles / switches
			int distance2Corral = dm.distance(cp.x, cp.y);
			boolean cornered = map.isCorner(cp.x, cp.y);

			if ((distance2Corral == 0) || (cornered==true))
				continue;
			
			// ignore cows that are too old
			if (cowTime < maxCowTime)
				continue;
			
			// does a cluster exists where this cow can be added to?
			for (CowCluster clu : cList) {
				
				// check distance along x,y axis against cluster cow 0				
				if ( clu.cows.get(0).HDistance( cp ) > cDist )
					continue;
				
				// check if we can get a path between that cow,
				// or if the cow is cornered.
				// because we already checked distances along
				// x and y, we can query the distance map if
				// that cow is a good option
				CowPosition cow0 = clu.cows.get(0);
				int distance0 = dm.distance(cow0.x, cow0.y);
				int absDist = Math.abs( distance0 - distance2Corral);
				if (absDist > (cDist+cDist/2))
					continue;
				
				// now try adding the cow
				if (clu.tryAdd(cp, cDist)) {
					clusterFound = true;
					break;
				}
			}
			
			if (!clusterFound) {
				cList.add( new CowCluster( new CowPosition(l)));
			}
		}
				
		return cList;
	}

	/**
	 * checks wether unpassable obstacles (agents, walls) are between two cows
	 * (we need a clean path between the two cows)
	 */
	protected boolean checkNoObstacles(CowPosition c1, CowPosition c2, GridMap m, int maxDist) {
		boolean ret =  false;
		try {
			int cDist = pathfinder.getPath(m, c1.x, c1.y, c2.x, c2.y, false).length;
			if (cDist <= maxDist && cDist > 0)
				ret = true;
		} catch (Exception e) {
			
		}
		return ret;
	}

		
}
